@@ -1,5 +1,6 @@
 # -*- coding=utf-8 -*-
 
+import atexit
 import contextlib
 import copy
 import functools
@@ -8,33 +9,71 @@ import os
 import attr
 import packaging.markers
 import packaging.version
+import pip_shims.shims
 import requests
-
-from first import first
 from packaging.utils import canonicalize_name
-
-from pip_shims import (
-    FormatControl, InstallRequirement, PackageFinder, RequirementPreparer,
-    RequirementSet, RequirementTracker, Resolver, WheelCache, pip_version
-)
-from vistir.compat import JSONDecodeError, TemporaryDirectory, fs_str
+from vistir.compat import JSONDecodeError, fs_str
 from vistir.contextmanagers import cd, temp_environ
-from vistir.misc import partialclass
+from vistir.path import create_tracked_tempdir
 
-from ..utils import get_pip_command, prepare_pip_source_args, _ensure_dir
+from ..environment import MYPY_RUNNING
+from ..utils import _ensure_dir, prepare_pip_source_args
 from .cache import CACHE_DIR, DependencyCache
+from .setup_info import SetupInfo
 from .utils import (
-    clean_requires_python, fix_requires_python_marker, format_requirement,
-    full_groupby, is_pinned_requirement, key_from_ireq,
-    make_install_requirement, name_from_req, version_from_ireq
+    clean_requires_python,
+    fix_requires_python_marker,
+    format_requirement,
+    full_groupby,
+    is_pinned_requirement,
+    key_from_ireq,
+    make_install_requirement,
+    name_from_req,
+    version_from_ireq,
 )
 
+try:
+    from contextlib import ExitStack
+except ImportError:
+    from contextlib2 import ExitStack
+
+if MYPY_RUNNING:
+    from typing import (
+        Any,
+        Dict,
+        List,
+        Generator,
+        Optional,
+        Union,
+        Tuple,
+        TypeVar,
+        Text,
+        Set,
+    )
+    from pip_shims.shims import (
+        InstallRequirement,
+        InstallationCandidate,
+        PackageFinder,
+        Command,
+    )
+    from packaging.requirements import Requirement as PackagingRequirement
+
+    TRequirement = TypeVar("TRequirement")
+    RequirementType = TypeVar(
+        "RequirementType", covariant=True, bound=PackagingRequirement
+    )
+    MarkerType = TypeVar("MarkerType", covariant=True, bound=Marker)
+    STRING_TYPE = Union[str, bytes, Text]
+    S = TypeVar("S", bytes, str, Text)
+
 
 PKGS_DOWNLOAD_DIR = fs_str(os.path.join(CACHE_DIR, "pkgs"))
 WHEEL_DOWNLOAD_DIR = fs_str(os.path.join(CACHE_DIR, "wheels"))
 
 DEPENDENCY_CACHE = DependencyCache()
-WHEEL_CACHE = WheelCache(CACHE_DIR, FormatControl(set(), set()))
+WHEEL_CACHE = pip_shims.shims.WheelCache(
+    CACHE_DIR, pip_shims.shims.FormatControl(set(), set())
+)
 
 
 def _get_filtered_versions(ireq, versions, prereleases):
@@ -42,6 +81,7 @@ def _get_filtered_versions(ireq, versions, prereleases):
 
 
 def find_all_matches(finder, ireq, pre=False):
+    # type: (PackageFinder, InstallRequirement, bool) -> List[InstallationCandidate]
     """Find all matching dependencies using the supplied finder and the
     given ireq.
 
@@ -52,6 +92,7 @@ def find_all_matches(finder, ireq, pre=False):
     :return: A list of matching candidates.
     :rtype: list[:class:`~pip._internal.index.InstallationCandidate`]
     """
+
     candidates = clean_requires_python(finder.find_all_candidates(ireq.name))
     versions = {candidate.version for candidate in candidates}
     allowed_versions = _get_filtered_versions(ireq, versions, pre)
@@ -61,9 +102,18 @@ def find_all_matches(finder, ireq, pre=False):
     return candidates
 
 
+def get_pip_command():
+    # type: () -> Command
+    # Use pip's parser for pip.conf management and defaults.
+    # General options (find_links, index_url, extra_index_url, trusted_host,
+    # and pre) are defered to pip.
+    pip_command = pip_shims.shims.InstallCommand()
+    return pip_command
+
+
 @attr.s
 class AbstractDependency(object):
-    name = attr.ib()
+    name = attr.ib()  # type: STRING_TYPE
     specifiers = attr.ib()
     markers = attr.ib()
     candidates = attr.ib()
@@ -94,9 +144,9 @@ class AbstractDependency(object):
         :rtype: set(str)
         """
 
-        if len(self.candidates) == 1 and first(self.candidates).editable:
+        if len(self.candidates) == 1 and next(iter(self.candidates)).editable:
             return self
-        elif len(other.candidates) == 1 and first(other.candidates).editable:
+        elif len(other.candidates) == 1 and next(iter(other.candidates)).editable:
             return other
         return self.version_set & other.version_set
 
@@ -113,15 +163,19 @@ class AbstractDependency(object):
 
         from .requirements import Requirement
 
-        if len(self.candidates) == 1 and first(self.candidates).editable:
+        if len(self.candidates) == 1 and next(iter(self.candidates)).editable:
             return self
-        elif len(other.candidates) == 1 and first(other.candidates).editable:
+        elif len(other.candidates) == 1 and next(iter(other.candidates)).editable:
             return other
         new_specifiers = self.specifiers & other.specifiers
-        markers = set(self.markers,) if self.markers else set()
+        markers = set(self.markers) if self.markers else set()
         if other.markers:
             markers.add(other.markers)
-        new_markers = packaging.markers.Marker(" or ".join(str(m) for m in sorted(markers)))
+        new_markers = None
+        if markers:
+            new_markers = packaging.markers.Marker(
+                " or ".join(str(m) for m in sorted(markers))
+            )
         new_ireq = copy.deepcopy(self.requirement.ireq)
         new_ireq.req.specifier = new_specifiers
         new_ireq.req.marker = new_markers
@@ -147,7 +201,7 @@ class AbstractDependency(object):
             requirement=new_requirement,
             parent=self.parent,
             dep_dict=dep_dict,
-            finder=self.finder
+            finder=self.finder,
         )
 
     def get_deps(self, candidate):
@@ -164,7 +218,7 @@ class AbstractDependency(object):
             from .requirements import Requirement
 
             req = Requirement.from_line(key)
-            req.merge_markers(self.markers)
+            req = req.merge_markers(self.markers)
             self.dep_dict[key] = req.get_abstract_dependencies()
         return self.dep_dict[key]
 
@@ -185,18 +239,23 @@ class AbstractDependency(object):
         extras = requirement.ireq.extras
         is_pinned = is_pinned_requirement(requirement.ireq)
         is_constraint = bool(parent)
-        finder = get_finder(sources=None)
+        _, finder = get_finder(sources=None)
         candidates = []
         if not is_pinned and not requirement.editable:
             for r in requirement.find_all_matches(finder=finder):
                 req = make_install_requirement(
-                    name, r.version, extras=extras, markers=markers, constraint=is_constraint,
+                    name,
+                    r.version,
+                    extras=extras,
+                    markers=markers,
+                    constraint=is_constraint,
                 )
-                req.req.link = r.location
+                req.req.link = getattr(r, "location", getattr(r, "link", None))
                 req.parent = parent
                 candidates.append(req)
                 candidates = sorted(
-                    set(candidates), key=lambda k: packaging.version.parse(version_from_ireq(k)),
+                    set(candidates),
+                    key=lambda k: packaging.version.parse(version_from_ireq(k)),
                 )
         else:
             candidates = [requirement.ireq]
@@ -238,10 +297,8 @@ def get_abstract_dependencies(reqs, sources=None, parent=None):
     from .requirements import Requirement
 
     for req in reqs:
-        if isinstance(req, InstallRequirement):
-            requirement = Requirement.from_line(
-                "{0}{1}".format(req.name, req.specifier)
-            )
+        if isinstance(req, pip_shims.shims.InstallRequirement):
+            requirement = Requirement.from_line("{0}{1}".format(req.name, req.specifier))
             if req.link:
                 requirement.req.link = req.link
                 requirement.markers = req.markers
@@ -258,6 +315,7 @@ def get_abstract_dependencies(reqs, sources=None, parent=None):
 
 
 def get_dependencies(ireq, sources=None, parent=None):
+    # type: (Union[InstallRequirement, InstallationCandidate], Optional[List[Dict[S, Union[S, bool]]]], Optional[AbstractDependency]) -> Set[S, ...]
     """Get all dependencies for a given install requirement.
 
     :param ireq: A single InstallRequirement
@@ -269,25 +327,27 @@ def get_dependencies(ireq, sources=None, parent=None):
     :return: A set of dependency lines for generating new InstallRequirements.
     :rtype: set(str)
     """
-    if not isinstance(ireq, InstallRequirement):
-        name = getattr(
-            ireq, "project_name",
-            getattr(ireq, "project", ireq.name),
-        )
-        version = getattr(ireq, "version")
-        ireq = InstallRequirement.from_line("{0}=={1}".format(name, version))
+    if not isinstance(ireq, pip_shims.shims.InstallRequirement):
+        name = getattr(ireq, "project_name", getattr(ireq, "project", ireq.name))
+        version = getattr(ireq, "version", None)
+        if not version:
+            ireq = pip_shims.shims.InstallRequirement.from_line("{0}".format(name))
+        else:
+            ireq = pip_shims.shims.InstallRequirement.from_line(
+                "{0}=={1}".format(name, version)
+            )
     pip_options = get_pip_options(sources=sources)
     getters = [
         get_dependencies_from_cache,
         get_dependencies_from_wheel_cache,
         get_dependencies_from_json,
-        functools.partial(get_dependencies_from_index, pip_options=pip_options)
+        functools.partial(get_dependencies_from_index, pip_options=pip_options),
     ]
     for getter in getters:
         deps = getter(ireq)
         if deps is not None:
             return deps
-    raise RuntimeError('failed to get dependencies for {}'.format(ireq))
+    raise RuntimeError("failed to get dependencies for {}".format(ireq))
 
 
 def get_dependencies_from_wheel_cache(ireq):
@@ -333,18 +393,22 @@ def get_dependencies_from_json(ireq):
         return
 
     session = requests.session()
+    atexit.register(session.close)
     version = str(ireq.req.specifier).lstrip("=")
 
     def gen(ireq):
         info = None
-        info = session.get(
-            "https://pypi.org/pypi/{0}/{1}/json".format(ireq.req.name, version)
-        ).json()["info"]
+        try:
+            info = session.get(
+                "https://pypi.org/pypi/{0}/{1}/json".format(ireq.req.name, version)
+            ).json()["info"]
+        finally:
+            session.close()
         requires_dist = info.get("requires_dist", info.get("requires"))
-        if not requires_dist:   # The API can return None for this.
+        if not requires_dist:  # The API can return None for this.
             return
         for requires in requires_dist:
-            i = InstallRequirement.from_line(requires)
+            i = pip_shims.shims.InstallRequirement.from_line(requires)
             # See above, we don't handle requirements with extras.
             if not _marker_contains_extra(i):
                 yield format_requirement(i)
@@ -379,12 +443,12 @@ def get_dependencies_from_cache(ireq):
     try:
         broken = False
         for line in cached:
-            dep_ireq = InstallRequirement.from_line(line)
+            dep_ireq = pip_shims.shims.InstallRequirement.from_line(line)
             name = canonicalize_name(dep_ireq.name)
             if _marker_contains_extra(dep_ireq):
-                broken = True   # The "extra =" marker breaks everything.
+                broken = True  # The "extra =" marker breaks everything.
             elif name == canonicalize_name(ireq.name):
-                broken = True   # A package cannot depend on itself.
+                broken = True  # A package cannot depend on itself.
             if broken:
                 break
     except Exception:
@@ -398,7 +462,7 @@ def get_dependencies_from_cache(ireq):
 
 
 def is_python(section):
-    return section.startswith('[') and ':' in section
+    return section.startswith("[") and ":" in section
 
 
 def get_dependencies_from_index(dep, sources=None, pip_options=None, wheel_cache=None):
@@ -412,85 +476,23 @@ def get_dependencies_from_index(dep, sources=None, pip_options=None, wheel_cache
     :rtype: set(str) or None
     """
 
-    finder = get_finder(sources=sources, pip_options=pip_options)
+    session, finder = get_finder(sources=sources, pip_options=pip_options)
     if not wheel_cache:
         wheel_cache = WHEEL_CACHE
     dep.is_direct = True
-    reqset = RequirementSet()
-    reqset.add_requirement(dep)
     requirements = None
     setup_requires = {}
-    with temp_environ(), start_resolver(finder=finder, wheel_cache=wheel_cache) as resolver:
-        os.environ['PIP_EXISTS_ACTION'] = 'i'
-        dist = None
+    with temp_environ():
+        os.environ["PIP_EXISTS_ACTION"] = "i"
         if dep.editable and not dep.prepared and not dep.req:
-            with cd(dep.setup_py_dir):
-                from setuptools.dist import distutils
-                try:
-                    dist = distutils.core.run_setup(dep.setup_py)
-                except (ImportError, TypeError, AttributeError):
-                    dist = None
-                else:
-                    setup_requires[dist.get_name()] = dist.setup_requires
-                if not dist:
-                    try:
-                        dist = dep.get_dist()
-                    except (TypeError, ValueError, AttributeError):
-                        pass
-                    else:
-                        setup_requires[dist.get_name()] = dist.setup_requires
-        resolver.require_hashes = False
-        try:
-            results = resolver._resolve_one(reqset, dep)
-        except Exception:
-            # FIXME: Needs to bubble the exception somehow to the user.
-            results = []
-        finally:
-            try:
-                wheel_cache.cleanup()
-            except AttributeError:
-                pass
-        resolver_requires_python = getattr(resolver, "requires_python", None)
-        requires_python = getattr(reqset, "requires_python", resolver_requires_python)
-        if requires_python:
-            add_marker = fix_requires_python_marker(requires_python)
-            reqset.remove(dep)
-            if dep.req.marker:
-                dep.req.marker._markers.extend(['and',].extend(add_marker._markers))
-            else:
-                dep.req.marker = add_marker
-            reqset.add(dep)
-        requirements = set()
-        for r in results:
-            if requires_python:
-                if r.req.marker:
-                    r.req.marker._markers.extend(['and',].extend(add_marker._markers))
-                else:
-                    r.req.marker = add_marker
-            requirements.add(format_requirement(r))
-        for section in setup_requires:
-            python_version = section
-            not_python = not is_python(section)
-
-            # This is for cleaning up :extras: formatted markers
-            # by adding them to the results of the resolver
-            # since any such extra would have been returned as a result anyway
-            for value in setup_requires[section]:
-
-                # This is a marker.
-                if is_python(section):
-                    python_version = value[1:-1]
-                else:
-                    not_python = True
-
-                if ':' not in value and not_python:
-                    try:
-                        requirement_str = "{0}{1}".format(value, python_version).replace(":", ";")
-                        requirements.add(format_requirement(make_install_requirement(requirement_str).ireq))
-                    # Anything could go wrong here -- can't be too careful.
-                    except Exception:
-                        pass
-
+            setup_info = SetupInfo.from_ireq(dep)
+            results = setup_info.get_info()
+            setup_requires.update(results["setup_requires"])
+            requirements = set(results["requires"].values())
+        else:
+            results = pip_shims.shims.resolve(dep)
+            requirements = [v for v in results.values() if v.name != dep.name]
+        requirements = set([format_requirement(r) for r in requirements])
     if not dep.editable and is_pinned_requirement(dep) and requirements is not None:
         DEPENDENCY_CACHE[dep] = list(requirements)
     return requirements
@@ -511,9 +513,7 @@ def get_pip_options(args=[], sources=None, pip_command=None):
     if not pip_command:
         pip_command = get_pip_command()
     if not sources:
-        sources = [
-            {"url": "https://pypi.org/simple", "name": "pypi", "verify_ssl": True}
-        ]
+        sources = [{"url": "https://pypi.org/simple", "name": "pypi", "verify_ssl": True}]
     _ensure_dir(CACHE_DIR)
     pip_args = args
     pip_args = prepare_pip_source_args(sources, pip_args)
@@ -523,6 +523,7 @@ def get_pip_options(args=[], sources=None, pip_command=None):
 
 
 def get_finder(sources=None, pip_command=None, pip_options=None):
+    # type: (List[Dict[S, Union[S, bool]]], Optional[Command], Any) -> PackageFinder
     """Get a package finder for looking up candidates to install
 
     :param sources: A list of pipfile-formatted sources, defaults to None
@@ -536,39 +537,38 @@ def get_finder(sources=None, pip_command=None, pip_options=None):
     """
 
     if not pip_command:
-        pip_command = get_pip_command()
+        pip_command = pip_shims.shims.InstallCommand()
     if not sources:
-        sources = [
-            {"url": "https://pypi.org/simple", "name": "pypi", "verify_ssl": True}
-        ]
+        sources = [{"url": "https://pypi.org/simple", "name": "pypi", "verify_ssl": True}]
     if not pip_options:
         pip_options = get_pip_options(sources=sources, pip_command=pip_command)
     session = pip_command._build_session(pip_options)
-    finder = PackageFinder(
-        find_links=[],
-        index_urls=[s.get("url") for s in sources],
-        trusted_hosts=[],
-        allow_all_prereleases=pip_options.pre,
-        session=session,
+    atexit.register(session.close)
+    finder = pip_shims.shims.get_package_finder(
+        pip_shims.shims.InstallCommand(), options=pip_options, session=session
     )
-    return finder
+    return session, finder
 
 
 @contextlib.contextmanager
-def start_resolver(finder=None, wheel_cache=None):
+def start_resolver(finder=None, session=None, wheel_cache=None):
     """Context manager to produce a resolver.
 
     :param finder: A package finder to use for searching the index
     :type finder: :class:`~pip._internal.index.PackageFinder`
+    :param :class:`~requests.Session` session: A session instance
+    :param :class:`~pip._internal.cache.WheelCache` wheel_cache: A pip WheelCache instance
     :return: A 3-tuple of finder, preparer, resolver
     :rtype: (:class:`~pip._internal.operations.prepare.RequirementPreparer`, :class:`~pip._internal.resolve.Resolver`)
     """
 
     pip_command = get_pip_command()
     pip_options = get_pip_options(pip_command=pip_command)
-
+    session = None
     if not finder:
-        finder = get_finder(pip_command=pip_command, pip_options=pip_options)
+        session, finder = get_finder(pip_command=pip_command, pip_options=pip_options)
+    if not session:
+        session = pip_command._build_session(pip_options)
 
     if not wheel_cache:
         wheel_cache = WHEEL_CACHE
@@ -577,37 +577,43 @@ def start_resolver(finder=None, wheel_cache=None):
     download_dir = PKGS_DOWNLOAD_DIR
     _ensure_dir(download_dir)
 
-    _build_dir = TemporaryDirectory(fs_str("build"))
-    _source_dir = TemporaryDirectory(fs_str("source"))
-    preparer = partialclass(
-        RequirementPreparer,
-        build_dir=_build_dir.name,
-        src_dir=_source_dir.name,
-        download_dir=download_dir,
-        wheel_download_dir=WHEEL_DOWNLOAD_DIR,
-        progress_bar="off",
-        build_isolation=False,
-    )
-    resolver = partialclass(
-        Resolver,
-        finder=finder,
-        session=finder.session,
-        upgrade_strategy="to-satisfy-only",
-        force_reinstall=True,
-        ignore_dependencies=False,
-        ignore_requires_python=True,
-        ignore_installed=True,
-        isolated=False,
-        wheel_cache=wheel_cache,
-        use_user_site=False,
-    )
-    if packaging.version.parse(pip_version) >= packaging.version.parse('18'):
-        with RequirementTracker() as req_tracker:
-            preparer = preparer(req_tracker=req_tracker)
-            yield resolver(preparer=preparer)
-    else:
-        preparer = preparer()
-        yield resolver(preparer=preparer)
+    _build_dir = create_tracked_tempdir(fs_str("build"))
+    _source_dir = create_tracked_tempdir(fs_str("source"))
+    try:
+        with ExitStack() as ctx:
+            ctx.enter_context(pip_shims.shims.global_tempdir_manager())
+            preparer = ctx.enter_context(
+                pip_shims.shims.make_preparer(
+                    options=pip_options,
+                    finder=finder,
+                    session=session,
+                    build_dir=_build_dir,
+                    src_dir=_source_dir,
+                    download_dir=download_dir,
+                    wheel_download_dir=WHEEL_DOWNLOAD_DIR,
+                    progress_bar="off",
+                    build_isolation=False,
+                    install_cmd=pip_command,
+                )
+            )
+            resolver = pip_shims.shims.get_resolver(
+                finder=finder,
+                ignore_dependencies=False,
+                ignore_requires_python=True,
+                preparer=preparer,
+                session=session,
+                options=pip_options,
+                install_cmd=pip_command,
+                wheel_cache=wheel_cache,
+                force_reinstall=True,
+                ignore_installed=True,
+                upgrade_strategy="to-satisfy-only",
+                isolated=False,
+                use_user_site=False,
+            )
+            yield resolver
+    finally:
+        session.close()
 
 
 def get_grouped_dependencies(constraints):
@@ -617,10 +623,10 @@ def get_grouped_dependencies(constraints):
     # then we take the loose match (which _is_ flexible) and start moving backwards in
     # versions by popping them off of a stack and checking for the conflicting package
     for _, ireqs in full_groupby(constraints, key=key_from_ireq):
-        ireqs = list(ireqs)
-        editable_ireq = first(ireqs, key=lambda ireq: ireq.editable)
+        ireqs = sorted(ireqs, key=lambda ireq: ireq.editable)
+        editable_ireq = next(iter(ireq for ireq in ireqs if ireq.editable), None)
         if editable_ireq:
-            yield editable_ireq  # ignore all the other specs: the editable one is the one that counts
+            yield editable_ireq  # only the editable match mattters, ignore all others
             continue
         ireqs = iter(ireqs)
         # deepcopy the accumulator so as to not modify the self.our_constraints invariant
